//! Actual demultiplexing logic: match input barcode module against reference.

use std::{
    cmp::{self, Ordering},
    collections::HashMap,
    iter,
};

use arrayvec::ArrayVec;

use crate::config::{self, Distance, MAX_MODULE_SIZE};

//--------------------------------------------------------------------------------------------------
// Stateful determination of the closest module(s)
// within the reference, given a sequence of the same size.
// Memoize results for performance.

// The closest module, with respect to Hamming distance,
// (i_module, distance) unless there are several matches.
type Closest = Option<(usize, usize)>;

// Local configuration.
pub(crate) struct Config {
    pub(crate) module_size: usize,
    pub(crate) n_modules: u64,
    pub(crate) distance: Distance,
}

pub(crate) struct Reference {
    cf: Config,
    /// All modules with same size, contiguously ordered.
    base: Vec<u8>,
    /// Memoization cache.
    cache: HashMap<Vec<u8>, Closest>,
    /// Useful to monitor memoization.
    pub(crate) n_queries: u64,
    pub(crate) n_hits: u64,
}

impl Reference {
    pub(crate) fn new(base: Vec<u8>, cf: Config) -> Self {
        Self {
            base,
            cf,
            cache: HashMap::new(),
            n_queries: 0,
            n_hits: 0,
        }
    }

    /// Exposed, memoized version.
    pub(crate) fn closest(&mut self, query: &[u8]) -> Option<(usize, usize)> {
        self.n_queries += 1;
        if let Some(res) = self.cache.get(query) {
            self.n_hits += 1;
            *res
        } else {
            let res = self.raw_closest(query);
            self.cache.insert(query.into(), res);
            res
        }
    }

    /// Internal, actual matching function.
    fn raw_closest(&self, query: &[u8]) -> Option<(usize, usize)> {
        use Distance as D;
        use Ordering as O;
        let s = self.cf.module_size;
        let nb = self.cf.n_modules;
        let nb = usize::try_from(nb)
            .unwrap_or_else(|e| panic!("To many barcode modules ({nb} to fit on local usize: {e}"));
        // Search for an *unique* reference with smallest distance.
        let mut min = s + 1; // ≈ ∞
        let mut found = None; // Raise on new min found, lower on duplicate min found.
        for i_module in 0..nb {
            // Extract reference.
            let start = i_module * s;
            let end = start + s;
            let module = &self.base[start..end];
            // TODO: profile: is this predictible enough to not require monomorphisation?
            let distance = match self.cf.distance {
                D::Hamming => hamming(query, module, min),
                D::Levenshtein => levenshtein(query, module),
            };
            match distance.cmp(&min) {
                O::Less => {
                    min = distance;
                    found = Some((i_module, min));
                }
                O::Equal => {
                    found = None; // Not unique.
                }
                O::Greater => {}
            }
        }
        found
    }
}

// Calculate Hamming distance
// assuming both inputs have the same size.
// Return early as soon as the threshold is overshot.
fn hamming(query: &[u8], reference: &[u8], threshold: usize) -> usize {
    let mut distance = 0;
    for (b, r) in query.iter().zip(reference.iter()) {
        distance += usize::from(b != r);
        if distance > threshold {
            break; // No need to calculate further.
        }
    }
    distance
}

// Calculate levenshtein distance up to the given threshold,
// assuming both inputs have the same size
// and this size is small enough to fit on the stack.
fn levenshtein(query: &[u8], reference: &[u8]) -> usize {
    const M: usize = MAX_MODULE_SIZE + 1;
    type Row = ArrayVec<usize, M>;
    // Classic row-major matrix implementation,
    // but only store the two focal rows.
    let n = reference.len();
    let new_row = |i| -> Row { iter::once(i).chain(iter::repeat(0)).take(n + 1).collect() };
    let mut previous: Row = (0..=n).collect();
    let mut current = new_row(1);
    for (q, i) in query.iter().zip(1..) {
        for (r, j) in reference.iter().zip(1..) {
            let deletion = previous[j] + 1;
            let insertion = current[j - 1] + 1;
            let substitution = previous[j - 1] + usize::from(q != r);
            current[j] = cmp::min(cmp::min(deletion, insertion), substitution);
        }
        if i < n {
            previous = current;
            current = new_row(i + 1);
        } else {
            return current[n];
        }
    }
    unreachable!()
}

impl From<&config::Config> for Config {
    fn from(cf: &config::Config) -> Self {
        let config::Config {
            n_modules,
            module_size,
            distance,
            ..
        } = *cf;
        Self {
            module_size,
            n_modules,
            distance,
        }
    }
}

//==================================================================================================

#[cfg(test)]
mod tests {
    use std::io::BufReader;

    use dedent::dedent;

    use super::*;
    use crate::{parse::modules, ContextualizedReader};

    #[test]
    fn match_module() {
        let cf = Config {
            n_modules: 4,
            module_size: 6,
            distance: Distance::Hamming,
        };
        let modules = dedent!(
            r"
            A01	ACGGAA
            A02	CCAACA
            A03	AGATCG
            A04	TTCTCC
        "
        );
        let reader = BufReader::new(modules.as_bytes());
        let reader = ContextualizedReader::new("<string>".into(), reader);
        let modules = modules::parse(reader, b'A', &cf).unwrap();
        let mut r = Reference::new(modules, cf);
        // Test thrice to trial cache.
        for _ in 0..3 {
            assert_eq!(r.closest(b"AGATCG"), Some((2, 0)));
            assert_eq!(r.closest(b"AGATGG"), Some((2, 1)));
            assert_eq!(r.closest(b"AGATGC"), Some((2, 2)));
            assert_eq!(r.closest(b"ACTAGA"), None);
        }
    }

    #[test]
    fn hamming_distance() {
        assert_eq!(hamming(b"AAAAAA", b"AATTAA", 6), 2);
        assert_eq!(hamming(b"AAAAAA", b"AATTAA", 2), 2);
        assert_eq!(hamming(b"AAAAAA", b"AATTAA", 1), 2);
        assert_eq!(hamming(b"AAAAAA", b"AATTAA", 0), 1);

        assert_eq!(hamming(b"ATGCAC", b"ATGCAC", 6), 0);
        assert_eq!(hamming(b"ATGCAC", b"ATGCAC", 6), 0);

        assert_eq!(hamming(b"GGGGGG", b"CCCCCC", 6), 6);
        assert_eq!(hamming(b"GGGGGG", b"CCCCCC", 0), 1);
    }

    #[test]
    fn levenshtein_distance() {
        assert_eq!(levenshtein(b"AAAAAA", b"AAAAAA"), 0);
        assert_eq!(levenshtein(b"AATAAA", b"AAAAAA"), 1);
        assert_eq!(levenshtein(b"AATAAT", b"AAAAAA"), 2);
        assert_eq!(levenshtein(b"012345", b"123456"), 2);
        assert_eq!(levenshtein(b"123456", b"012345"), 2);
        assert_eq!(levenshtein(b"12X345", b"012345"), 2);
        assert_eq!(levenshtein(b"01X234", b"012345"), 2);
    }
}
