//! This thread keeps track of idle writer threads
//! and of the largest paired buffers ready to be written
//! in terms of number of fastq blocks,
//! so as to dispatch the largest jobs to idle writers until termination.

use std::{
    collections::HashMap,
    io::Write,
    mem, panic,
    sync::{
        mpsc::{Receiver, TryRecvError},
        Arc, Mutex,
    },
};

use priority_queue::PriorityQueue;
use snafu::Snafu;

use crate::{writer, ContextualizedWriter};

/// { sample code ↦ ([R1, R2], number of blocks ready to output) }
pub(crate) type BufferMap = HashMap<usize, ([Vec<u8>; 2], usize)>;

pub(crate) struct Inner<W: Write> {
    /// Writer threads to dispatch jobs to.
    pub(crate) workers: Vec<writer::Outer<W>>,
    /// Writers to sample files to zip/write buffers to.
    pub(crate) all_writers: HashMap<usize, [ContextualizedWriter<W>; 2]>,
    /// Buffers filled by the main thread with the data to be written.
    pub(crate) buffers: Arc<Mutex<BufferMap>>,
    /// { Sample code ↦  number of fastq blocks ready for writing }
    pub(crate) queue: Arc<Mutex<PriorityQueue<usize, usize>>>,
    /// Receive writer thread outputs over this channel.
    pub(crate) writers_output: Receiver<writer::Output<W>>,
    /// Receive termination signal over this channel.
    pub(crate) termination: Receiver<()>,
    /// Count the number of output allocations created and reused,
    /// Send them back as return value of the whole thread.
    pub(crate) allocs_created: u64,
    pub(crate) allocs_reused: u64,
}

impl<W: Write> Inner<W> {
    #[allow(clippy::too_many_lines)] // Ongoing experiment.
    pub(crate) fn run(self) -> Result<(u64, u64), Error> {
        let Self {
            mut workers,
            mut all_writers, // Consumed during termination as they are flushed by workers.
            queue,
            buffers,
            mut allocs_created,
            mut allocs_reused,
            termination,
            writers_output,
        } = self;
        let mut idle: Vec<_> = (0..workers.len()).collect();
        let mut free_buffers = Vec::new(); // (for recycling allocations)
        let mut terminate = false; // Raise when no more input is expected.
        let mut peel = Vec::new(); // (temporary store top-priority jobs without available writers)
        'main: loop {
            // Pick an idle worker.
            let id = idle
                .pop()
                .expect("Should block until there are idle workers.");
            let thread = &mut workers[id];

            // Grab the largest buffer whose writer is not busy at the moment.
            let largest = 'a: {
                let cw = {
                    let mut lock = queue.lock().expect("poisoned queue?");
                    // PriorityQueue cannot peek into *several* first items.
                    // We need to consume it until we find a free writer,
                    // and then restore the items "peeled" during the search.
                    let mut found = None;
                    while let Some((code, nb)) = lock.pop() {
                        peel.push((code, nb));
                        if !terminate && nb == 0 {
                            break; // Only pick empty buffers for flushing on termination.
                        }
                        if let Some(writers) = all_writers.remove(&code) {
                            found = Some((code, writers));
                            break;
                        }
                    }
                    for (code, nb) in peel.drain(0..) {
                        lock.push(code, nb);
                    }
                    found
                };
                let Some((code, writers)) = cw else {
                    // All relevant writers are already busy. Skip ahead.
                    break 'a None;
                };
                // Retrieve corresponding buffers to send data to the worker.
                let mut buffers_lock = buffers.lock().expect("poisoned buffers?");
                let mut queue_lock = queue.lock().expect("poisoned queue?");
                let (buffers, n) = buffers_lock.get_mut(&code).unwrap();
                // Extract the data to send and replace with empty recycled slot.
                let mut send = if let Some(free) = free_buffers.pop() {
                    allocs_reused += 2;
                    free
                } else {
                    allocs_created += 2;
                    [Vec::new(), Vec::new()]
                };
                mem::swap(buffers, &mut send);
                let nb = *n;
                *n = 0;
                // Update queue.
                queue_lock.change_priority(&code, 0);
                drop(queue_lock);
                drop(buffers_lock);
                Some((code, writers, send, nb))
            };

            // Send this buffer + writer as a writing job to this thread.
            if let Some((sample_code, writers, buffers, n_blocks)) = largest {
                thread
                    .input
                    .send(writer::Input::Write {
                        sample_code,
                        buffers,
                        writers,
                        n_blocks,
                        flush: terminate,
                    })
                    .map_err(|e| Error::Message {
                        id,
                        erm: e.to_string(),
                    })?;
            } else {
                // Found no work to perform: this thread will not get to work.
                idle.push(id);
            }

            if !terminate {
                // Keep an eye on termination.
                match termination.try_recv() {
                    Err(TryRecvError::Empty) => {}
                    Ok(()) => {
                        // No more input is expected, but keep working
                        // until the queue is empty and all writers are consumed.
                        terminate = true;
                    }
                    Err(TryRecvError::Disconnected) => {
                        panic!("Main thread exited: dispatcher panic.");
                    }
                }
            }
            if terminate && all_writers.is_empty() {
                break 'main;
            }

            // Collect idle workers.
            let empty = idle.is_empty();
            let mut process = |message| match message {
                writer::Output::Written {
                    id,
                    sample_code,
                    buffers,
                    writers,
                } => {
                    idle.push(id);
                    all_writers.insert(sample_code, writers);
                    free_buffers.push(buffers);
                }
                writer::Output::Flushed { id, buffers } => {
                    idle.push(id);
                    free_buffers.push(buffers);
                    // Writers consumed by the worker.
                }
            };
            // If we have none left, block until there is at least one.
            if empty {
                for message in writers_output.iter().take(1) {
                    process(message);
                }
            }
            for message in writers_output.try_iter() {
                process(message);
            }
        }

        // Terminate.
        println!("Send termination signals to writer threads.");
        for (id, writer) in workers.iter_mut().enumerate() {
            writer
                .input
                .send(writer::Input::Terminate)
                .map_err(|e| Error::Message {
                    id,
                    erm: e.to_string(),
                })?;
        }

        println!("Wait for writers to terminate.");
        for (i, writer) in workers.into_iter().enumerate() {
            match writer.handle.join() {
                Ok(Ok(())) => {}
                Ok(Err(err)) => return Err(err.into()),
                Err(err) => {
                    eprintln!(
                        "Writer thread {i:02} unexpectedly exited: \
                                     propagate panic to main thread.",
                    );
                    panic::resume_unwind(err);
                }
            }
        }

        Ok((allocs_created, allocs_reused))
    }
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(transparent)]
    WriterThread { source: writer::Error },
    #[snafu(display(
        "Message could not be sent from dispatcher thread \
         to writer thread {id}:\n{erm}"
    ))]
    Message {
        id: usize,
        erm: String, // Rendered underlying error message.
    },
}
