//! Check and collect all reference modules from the given file.
//! References are returned as a single buffer,
//! easy to index because they have all been checked for being the same size.

use std::{io::Read, num::ParseIntError, str::Utf8Error};

use snafu::{ensure, OptionExt, ResultExt, Snafu};

use crate::{
    demux::Config,
    io::{self, Context},
    short::ful,
    ContextualizedReader,
};

pub(crate) fn parse<R: Read>(
    mut reader: ContextualizedReader<R>,
    letter: u8,
    cf: &Config,
) -> Result<Vec<u8>, Error> {
    let &Config {
        n_modules,
        module_size,
        distance: _,
    } = cf;

    let mut res = Vec::new();
    let mut line = Vec::new();
    while reader.read_until(b'\n', &mut line)? > 0 {
        let i_line = reader.i_line();
        if line.is_empty() && i_line == n_modules + 1 {
            continue; // Allow trailing empty line.
        }
        if *line.last().unwrap() == b'\n' {
            line.pop();
        }
        let code = extract_code(i_line, &line, letter, cf)
            .with_context(|_| ParseErr { context: &reader })?;
        res.extend_from_slice(code);
        reader.bump_line();
        line.clear();
    }
    let bs = module_size as u64;
    ensure!(
        res.len() as u64 == n_modules * bs,
        NLinesErr {
            actual: res.len() as u64 / bs,
            expected: n_modules,
        }
    );

    Ok(res)
}

fn extract_code<'l>(
    i_line: u64,
    line: &'l [u8],
    letter: u8,
    cf: &Config,
) -> Result<&'l [u8], ParseError> {
    let &Config { module_size, .. } = cf;
    let (id, code) = line
        .split_once(|&b| b == b'\t')
        .with_context(|| LineErr { line })?;

    // Parse id and check consistency with line number.
    let id = id
        .strip_prefix(&[letter])
        .with_context(|| IdPrefixErr { ref_id: letter, id })?;
    let i_id = std::str::from_utf8(id)
        .with_context(|_| IdUtf8Err { id })?
        .parse::<u64>()
        .with_context(|_| IdIntErr { id })?;
    ensure!(i_id == i_line, IdMismatchErr { id });

    // Check module size.
    let actual = code.len() as u64;
    let expected = module_size as u64;
    ensure!(
        actual == expected,
        ModuleLenErr {
            actual,
            expected,
            code,
        }
    );

    // Check module content.
    for (pos, &char) in code.iter().enumerate() {
        ensure!([b'A', b'T', b'G', b'C'].iter().any(|&base| base == char), {
            BaseErr {
                fault: char,
                pos,
                code,
            }
        });
    }

    Ok(code)
}

//==================================================================================================
#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(transparent)]
    Io { source: io::Error },
    #[snafu(display("{context}{source}"))]
    Parse {
        context: Context,
        source: ParseError,
    },
    #[snafu(display("Expected {expected} barcode modules, found {actual} instead."))]
    NLines { actual: u64, expected: u64 },
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum ParseError {
    #[snafu(display(
        "Could not parse as '<module_id><tab><module_sequence>':\n  {:?}",
        ful(line)
    ))]
    Line { line: Vec<u8> },
    #[snafu(display("Barcode module id does not start with {ref_id:?}: {:?}.", ful(id)))]
    IdPrefix { ref_id: char, id: Vec<u8> },
    #[snafu(display("Invalid utf8 in barcode module id index {:?}:\n{source}", ful(id)))]
    IdUtf8 { source: Utf8Error, id: Vec<u8> },
    #[snafu(display("Invalid barcode module id index: {:?}:\n{source}", ful(id)))]
    IdInt { source: ParseIntError, id: Vec<u8> },
    #[snafu(display(
        "The barcode module index differs from the line number: {:?}.",
        ful(id)
    ))]
    IdMismatch { id: Vec<u8> },
    #[snafu(display(
        "Expected a barcode module of size {expected}, got {actual} instead: {:?}.",
        ful(code)
    ))]
    ModuleLen {
        actual: u64,
        expected: u64,
        code: Vec<u8>,
    },
    #[snafu(display(
        "Invalid character in barcode module at position {pos}: {fault:?} in {:?}.",
        ful(code)
    ))]
    Base {
        fault: char,
        pos: usize,
        code: Vec<u8>,
    },
}
